/* init.c - init daemon
 * Time-stamp: <2004-04-13 njk>
 *  
 * (C) 2004 Nicholas J. Kain <njk@aerifal.cx>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/reboot.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <linux/kd.h>

#include "signals.h"

#define INIT_DIR     "/etc/init/"
#define BSCRIPT      "boot"
#define SDSCRIPT     "shutdown"
#define RBSCRIPT     "reboot"
#define POSCRIPT     "poweroff"
#define SUSCRIPT     "single"
#define SPATH        "/sbin/"
#define SSHELL       "sulogin"
#define KBRSCRIPT    "kbreq"

static volatile int pending_exit, pending_reboot, pending_reap, pending_kbreq,
	pending_poweroff, got_cont, dying;
static int gflags_single, gflags_emergency;

static void sighandler(int sig) {
	switch (sig) {
		case SIGTERM:
			pending_exit = 1;
			break;
		case SIGINT:
			pending_reboot = 1;
			break;
		case SIGCHLD:
			pending_reap = 1;
			break;
		case SIGWINCH:
			pending_kbreq = 1;
			break;
		case SIGUSR1:
			pending_poweroff = 1;
			break;
	}
}

static void sig_cont(int sig) {
	got_cont = 1;
}

static void sig_stop(int sig) {
	int sav_errno = errno;

	got_cont = 0;
	while (!got_cont) pause();
	got_cont = 0;
	errno = sav_errno;
}

static void fix_signals(void) {
  disable_signal(SIGPIPE);
  disable_signal(SIGUSR2);
  disable_signal(SIGTSTP);
  disable_signal(SIGTTIN);
  disable_signal(SIGHUP);
 
  hook_signal(SIGCHLD, sighandler, 0);   /* reap zombies */
  hook_signal(SIGINT, sighandler, 0);    /* ctrl-alt-del */
  hook_signal(SIGTERM, sighandler, 0);   /* death */
  hook_signal(SIGWINCH, sighandler, 0);  /* keyboard request */
  hook_signal(SIGUSR1, sighandler, 0);   /* power off request */
  hook_signal(SIGCONT, sig_cont, 0);
  hook_signal(SIGTSTP, sig_stop, 0);
  hook_signal(SIGSTOP, sig_stop, 0);
}

static void open_console(void) {
	int fd;
	if ( (fd = open("/dev/console", O_RDWR|O_NOCTTY)) >=0) {
		dup2(fd, 0);
		dup2(fd, 1);
		dup2(fd, 2);
		if (fd>2) close(fd);
	}
}

static void reave_zombies(void) {
	while(waitpid(-1, NULL, WNOHANG) > 0);
	pending_reap = 0;
}

static void do_exec(char *path, char *file, int shouldwait) {
	pid_t t;
	switch ((t = fork())) {
		case 0:
			setsid();
			open_console();
			tcsetpgrp(0, getpgrp());
			execl(path, file, NULL);
			break;
		case -1:
			break;
		default:
			if (shouldwait) while (waitpid(t, NULL, 0) != t);
			break;
	}
}

static void do_reboot(void) {
	pending_reboot = 0;
	dying = 1;
	do_exec(INIT_DIR RBSCRIPT, RBSCRIPT, 0);
}

static void do_exit(void) {
	pending_exit = 0;
	dying = 1;
	do_exec(INIT_DIR SDSCRIPT, SDSCRIPT, 0);
}

static void do_kbreq(void) {
	pending_kbreq = 0;
	do_exec(INIT_DIR KBRSCRIPT, KBRSCRIPT, 0);
}

static void do_poweroff(void) {
	pending_poweroff = 0;
	dying = 1;
	do_exec(INIT_DIR POSCRIPT, POSCRIPT, 0);
}

static void mundane_loop(void) {
	while (1) {
		if (pending_reap) reave_zombies();
		if (!dying) {
			if (pending_reboot) do_reboot();
			if (pending_exit) do_exit();
			if (pending_poweroff) do_poweroff();
		}
		if (pending_kbreq) do_kbreq();
		pause();
	}
}

static void spawn_emergency(void) {
	do_exec(SPATH SSHELL, SSHELL, 0);
	mundane_loop();
}

static void spawn_single(void) {
	do_exec(INIT_DIR SUSCRIPT, SUSCRIPT, 0);
	mundane_loop();
}

static void spawn_boot(void) {
	do_exec(INIT_DIR BSCRIPT, BSCRIPT, 1);
	mundane_loop();
}

static void do_init(void) {
	if (gflags_single) {
		spawn_single();
		return;
	}
	if (gflags_emergency) {
		spawn_emergency();
		return;
	}
	spawn_boot();
}

int main(int argc, char** argv) {
	int c, fd, fdash = 0;
	char *p = argv[1];
	
	while (p && *p) {
	    switch (*p) {
			case '-':
				fdash = 1;
				break;
	        case 'b':
				if (fdash)
		            gflags_emergency = 1;
	            break;
	        case 's':
				if (fdash)
		            gflags_single = 1; 
	            break;
			default:
				fdash = 0;
				break;
		}
		++p;
	}
	
	fix_signals();
	reboot(RB_DISABLE_CAD);

	if ((fd = open("/dev/console", O_RDWR|O_NOCTTY)) >= 0) {
		ioctl(fd, KDSIGACCEPT, SIGWINCH);
		close(fd);
	} else ioctl(0, KDSIGACCEPT, SIGWINCH);

	for (c=0; c<3; c++) {
		close(c);
		open("/dev/null", O_RDWR, 0);
	}
	
	do_init();
  
	exit(EXIT_FAILURE);
}

